/*  $Id: surface_target_opengl.cpp,v 1.10 2003/10/29 15:03:02 grumbel Exp $
**
**  ClanLib Game SDK
**  Copyright (C) 2003  The ClanLib Team
**  For a total list of contributers see the file CREDITS.
**
**  This library is free software; you can redistribute it and/or
**  modify it under the terms of the GNU Lesser General Public
**  License as published by the Free Software Foundation; either
**  version 2.1 of the License, or (at your option) any later version.
**
**  This library is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
**  Lesser General Public License for more details.
**
**  You should have received a copy of the GNU Lesser General Public
**  License along with this library; if not, write to the Free Software
**  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
*/

#include "Display/display_precomp.h"
#include "surface_target_opengl.h"
#include "gl_state_checker.h"
#include "graphic_context_opengl.h"
#ifdef WIN32
#include "WGL/display_window_opengl.h"
#else
#include "GLX/display_window_opengl.h"
#endif
#include "API/Display/pixel_buffer.h"
#include "API/Display/graphic_context.h"
#include "API/Display/pixel_format.h"
#include "API/Display/display.h"
#include "API/Core/System/error.h"
#include <Hermes/H_Conv.h>

#include <GL/gl.h>
#include <GL/glu.h>

#ifndef GL_BGR
#define GL_BGR 0x80E0
#endif
#ifndef GL_UNSIGNED_BYTE_3_3_2
#define GL_UNSIGNED_BYTE_3_3_2              0x8032
#endif
#ifndef GL_UNSIGNED_BYTE_3_3_2_EXT
#define GL_UNSIGNED_BYTE_3_3_2_EXT          0x8032
#endif
#ifndef GL_UNSIGNED_SHORT_4_4_4_4
#define GL_UNSIGNED_SHORT_4_4_4_4           0x8033
#endif
#ifndef GL_UNSIGNED_SHORT_4_4_4_4_EXT
#define GL_UNSIGNED_SHORT_4_4_4_4_EXT       0x8033
#endif
#ifndef GL_UNSIGNED_SHORT_5_5_5_1
#define GL_UNSIGNED_SHORT_5_5_5_1           0x8034
#endif
#ifndef GL_UNSIGNED_SHORT_5_5_5_1_EXT
#define GL_UNSIGNED_SHORT_5_5_5_1_EXT       0x8034
#endif
#ifndef GL_UNSIGNED_INT_8_8_8_8
#define GL_UNSIGNED_INT_8_8_8_8             0x8035
#endif
#ifndef GL_UNSIGNED_INT_8_8_8_8_EXT
#define GL_UNSIGNED_INT_8_8_8_8_EXT         0x8035
#endif
#ifndef GL_UNSIGNED_INT_10_10_10_2
#define GL_UNSIGNED_INT_10_10_10_2          0x8036
#endif
#ifndef GL_UNSIGNED_INT_10_10_10_2_EXT
#define GL_UNSIGNED_INT_10_10_10_2_EXT      0x8036
#endif
#ifndef GL_UNSIGNED_BYTE_2_3_3_REV
#define GL_UNSIGNED_BYTE_2_3_3_REV          0x8362
#endif
#ifndef GL_UNSIGNED_BYTE_2_3_3_REV_EXT
#define GL_UNSIGNED_BYTE_2_3_3_REV_EXT      0x8362
#endif
#ifndef GL_UNSIGNED_SHORT_5_6_5
#define GL_UNSIGNED_SHORT_5_6_5             0x8363
#endif
#ifndef GL_UNSIGNED_SHORT_5_6_5_EXT
#define GL_UNSIGNED_SHORT_5_6_5_EXT         0x8363
#endif
#ifndef GL_UNSIGNED_SHORT_5_6_5_REV
#define GL_UNSIGNED_SHORT_5_6_5_REV         0x8364
#endif
#ifndef GL_UNSIGNED_SHORT_5_6_5_REV_EXT
#define GL_UNSIGNED_SHORT_5_6_5_REV_EXT     0x8364
#endif
#ifndef GL_UNSIGNED_SHORT_4_4_4_4_REV
#define GL_UNSIGNED_SHORT_4_4_4_4_REV       0x8365
#endif
#ifndef GL_UNSIGNED_SHORT_4_4_4_4_REV_EXT
#define GL_UNSIGNED_SHORT_4_4_4_4_REV_EXT   0x8365
#endif
#ifndef GL_UNSIGNED_SHORT_1_5_5_5_REV
#define GL_UNSIGNED_SHORT_1_5_5_5_REV       0x8366
#endif
#ifndef GL_UNSIGNED_SHORT_1_5_5_5_REV_EXT
#define GL_UNSIGNED_SHORT_1_5_5_5_REV_EXT   0x8366
#endif
#ifndef GL_UNSIGNED_INT_8_8_8_8_REV
#define GL_UNSIGNED_INT_8_8_8_8_REV         0x8367
#endif
#ifndef GL_UNSIGNED_INT_8_8_8_8_REV_EXT
#define GL_UNSIGNED_INT_8_8_8_8_REV_EXT     0x8367
#endif
#ifndef GL_UNSIGNED_INT_2_10_10_10_REV
#define GL_UNSIGNED_INT_2_10_10_10_REV      0x8368
#endif
#ifndef GL_UNSIGNED_INT_2_10_10_10_REV_EXT
#define GL_UNSIGNED_INT_2_10_10_10_REV_EXT  0x8368
#endif

/////////////////////////////////////////////////////////////////////////////
// CL_Surface_Target_OpenGL construction:

CL_Surface_Target_OpenGL::CL_Surface_Target_OpenGL(
	CL_PixelBuffer *provider,
	bool delete_provider,
	CL_Surface::Hint hint)
{
	CL_GLStateChecker::from_gc(CL_Display::get_current_window()->get_gc())->flush();

	gc = 0;
#ifdef WIN32
	pbuffer_context = 0;
	pbuffer_dc = 0;

	if (wglCreatePbufferARB == 0)
	{
		wglCreatePbufferARB = (ptr_wglCreatePbufferARB) wglGetProcAddress("wglCreatePbufferARB");
		wglCreatePbufferARB = (ptr_wglCreatePbufferARB) wglGetProcAddress("wglCreatePbufferARB");
		wglGetPbufferDCARB = (ptr_wglGetPbufferDCARB) wglGetProcAddress("wglGetPbufferDCARB");
		wglReleasePbufferDCARB = (ptr_wglReleasePbufferDCARB) wglGetProcAddress("wglReleasePbufferDCARB");
		wglDestroyPbufferARB = (ptr_wglDestroyPbufferARB) wglGetProcAddress("wglDestroyPbufferARB");
		wglQueryPbufferARB = (ptr_wglQueryPbufferARB) wglGetProcAddress("wglQueryPbufferARB");
	}
#else
	// todo: add linux pbuffer init here.
#endif
	pbuffer_changed = false;
	texture_changed = true;

	// Convert provider to OpenGL texture format ABGR:
	CL_PixelFormat abgr_format;
	abgr_format.set_depth(32);
	abgr_format.enable_colorkey(false);
	abgr_format.set_alpha_mask(0xFF000000);
	abgr_format.set_blue_mask (0x00FF0000);
	abgr_format.set_green_mask(0x0000FF00);
	abgr_format.set_red_mask  (0x000000FF);

	provider->lock();

	screen_size.width = (float)provider->get_width();
	screen_size.height = (float)provider->get_height();
	
	texture_size.width = 1;
	texture_size.height = 1;
	while (texture_size.width < screen_size.width) texture_size.width *= 2;
	while (texture_size.height < screen_size.height) texture_size.height *= 2;

	CL_PixelBuffer *buffer = new CL_PixelBuffer(
		static_cast<int>(texture_size.width),
		static_cast<int>(texture_size.height),
		static_cast<int>(texture_size.width) * 4,
		abgr_format);

	provider->convert(
		buffer->get_data(), abgr_format,
		buffer->get_pitch(),
		CL_Rect(0, 0, static_cast<int>(screen_size.width), static_cast<int>(screen_size.height)),
		CL_Rect(0, 0, static_cast<int>(screen_size.width), static_cast<int>(screen_size.height)));

	provider->unlock();
	if (delete_provider) delete provider;

	buffer->lock();
	
	// Detect what level of alpha we require if application didnt tell us:
	if (hint == CL_Surface::hint_automatic)
	{
		bool found_colorkey = false;
		bool found_alpha = false;
	
		unsigned char *data = (unsigned char *) buffer->get_data();
		for (int y=0; y<screen_size.height; y++)
		{
			for (int x=0; x<screen_size.width; x++)
			{
				unsigned char alpha = data[x*4+3];
				if (alpha == 0) found_colorkey = true;
				else if (alpha != 255) found_alpha = true;
			}
			data += buffer->get_pitch();
		}
		
		if (found_alpha) hint = CL_Surface::hint_alpha;
		else if (found_colorkey) hint = CL_Surface::hint_colorkey;
		else hint = CL_Surface::hint_opaque;
	}
	
	// Figure out what is the optimal supported format for this card:
	GLint desired_format = GL_RGBA8;
	if (hint == CL_Surface::hint_opaque) desired_format = GL_RGB8;
	glTexImage2D(
		GL_PROXY_TEXTURE_2D,      // target
		0,                        // level
		desired_format,           // internalformat
		int(texture_size.width),  // width
		int(texture_size.height), // height
		0,                        // border
		GL_RGBA,                  // format
		GL_UNSIGNED_BYTE,         // type
		buffer->get_data());      // texels

	GLint obtained_format = desired_format;
	glGetTexLevelParameteriv(GL_PROXY_TEXTURE_2D, 0, GL_TEXTURE_INTERNAL_FORMAT, &obtained_format);
	switch (hint)
	{
	case CL_Surface::hint_alpha:
		break;
	
	case CL_Surface::hint_colorkey:
		// If we got 16 bit, try to get 5 bits for colors if possible.
		if (obtained_format == GL_RGBA4) desired_format = GL_RGB5_A1;
		break;
	
	case CL_Surface::hint_opaque:
		break;

	case CL_Surface::hint_automatic:
		break;
	}

	// Upload to OpenGL:
	glGenTextures(1, &handle);
	glBindTexture(GL_TEXTURE_2D, handle);

	glTexImage2D(
		GL_TEXTURE_2D,            // target
		0,                        // level
		desired_format,           // internalformat
		int(texture_size.width),  // width
		int(texture_size.height), // height
		0,                        // border
		GL_RGBA,                  // format
		GL_UNSIGNED_BYTE,         // type
		buffer->get_data());      // texels

	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
//	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
//	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
//	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
//	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
//	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_MIRRORED_REPEAT);
//	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_MIRRORED_REPEAT);
//	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_BORDER);
//	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_BORDER);

	buffer->unlock();
	delete buffer;
}

CL_Surface_Target_OpenGL::~CL_Surface_Target_OpenGL()
{
#ifdef WIN32
	if (pbuffer_context)
	{
		wglMakeCurrent(0, 0);
		wglDeleteContext(pbuffer_context);
	}
	if (pbuffer_dc) wglReleasePbufferDCARB(pbuffer, pbuffer_dc);
	if (pbuffer) wglDestroyPbufferARB(pbuffer);
#else
	// todo: add linux pbuffer deinit here.
#endif

	glDeleteTextures(1, &handle);
}

/////////////////////////////////////////////////////////////////////////////
// CL_Surface_Target_OpenGL attributes:

int CL_Surface_Target_OpenGL::get_width() const
{
	return (int)screen_size.width;
}
	
int CL_Surface_Target_OpenGL::get_height() const
{
	return (int)screen_size.height;
}

CL_PixelBuffer CL_Surface_Target_OpenGL::get_pixeldata()
{
	return CL_PixelBuffer();
}
	
CL_GraphicContext *CL_Surface_Target_OpenGL::get_gc()
{
	if (gc == 0)
	{
		// Create working pbuffer:
#ifdef WIN32
		if (wglCreatePbufferARB == 0)
		{
			throw CL_Error("WGL_ARB_pbuffer OpenGL extension not supported by this card");
		}

		int attribList[1] = { 0 };

		PIXELFORMATDESCRIPTOR pfd =
		{
			sizeof(PIXELFORMATDESCRIPTOR),  // size of this pfd 
			1,                              // version number
			// PFD_DRAW_TO_WINDOW |            // support window
			PFD_SUPPORT_OPENGL //|            // support OpenGL
			//PFD_DOUBLEBUFFER |              // double buffered
			//PFD_DEPTH_DONTCARE
			,             // do you care about a zbuffer?
			PFD_TYPE_RGBA,                  // RGBA type
			24,                             // 24-bit color depth
			0, 0, 0, 0, 0, 0,               // color bits ignored
			8,                              // no alpha buffer
			0,                              // shift bit ignored
			0,                              // no accumulation buffer
			0, 0, 0, 0,                     // accum bits ignored
			0,                              // z-buffer
			0,                              // no stencil buffer
			0,                              // no auxiliary buffer
			PFD_MAIN_PLANE,                 // main layer
			0,                              // reserved
			0, 0, 0                         // layer masks ignored
		};

		int pixelformat = ChoosePixelFormat(wglGetCurrentDC(), &pfd);

//		int pixelformat = GetPixelFormat(wglGetCurrentDC());

		pbuffer = wglCreatePbufferARB(
			wglGetCurrentDC(),
			pixelformat,
			int(screen_size.width),
			int(screen_size.height),
			attribList);
		pbuffer_dc = wglGetPbufferDCARB(pbuffer);
		pbuffer_context = wglCreateContext(pbuffer_dc);

		if (!CL_DisplayWindow_OpenGL::opengl_contexts.empty())
		{
			wglShareLists(CL_DisplayWindow_OpenGL::opengl_contexts.back(), pbuffer_context);
		}

		CL_GLStateChecker::set_context(pbuffer_dc, pbuffer_context);
#else
		// todo: create pbuffer in linux here.
#endif

		// Create GC interface for it:
		gc = new CL_GraphicContext(new CL_GraphicContext_OpenGL(this));

		// And setup the GC interface:
		gc->impl->setup_2d();

		// Download surface to pbuffer:
		pbuffer_changed = false;
		texture_changed = true;
		sync_pbuffer();
	}

	return gc;
}

/////////////////////////////////////////////////////////////////////////////
// CL_Surface_Target_OpenGL operations:

void CL_Surface_Target_OpenGL::set_pixeldata(
	const CL_Point &pos,
	const CL_Rect &src_rect,
	CL_PixelBuffer &pixeldata)
{
	CL_GLStateChecker::from_gc(CL_Display::get_current_window()->get_gc())->flush();

	// Go look if OpenGL already know this format:
	const CL_PixelFormat &src_format = pixeldata.get_format();
	
	int gl_format = GL_RGBA8;
	int gl_type = GL_UNSIGNED_BYTE;
	bool do_pb_convert = false;
	if (src_format == CL_PixelFormat::rgba8888)
	{
		gl_format = GL_RGBA;
		gl_type = GL_UNSIGNED_INT_8_8_8_8;
	}
	else if (src_format == CL_PixelFormat::rgb888)
	{
		gl_format = GL_BGR;
		gl_type = GL_UNSIGNED_BYTE;
	}
	else if (src_format == CL_PixelFormat::rgba4444)
	{
		gl_format = GL_RGBA;
		gl_type = GL_UNSIGNED_SHORT_4_4_4_4;
	}
	else if (src_format == CL_PixelFormat::rgb565)
	{
		gl_format = GL_RGB;
		gl_type = GL_UNSIGNED_SHORT_5_6_5;
	}
	else if (src_format == CL_PixelFormat::abgr8888)
	{
		gl_format = GL_RGBA;
		gl_type = GL_UNSIGNED_INT_8_8_8_8_REV;
	}
	else if (src_format == CL_PixelFormat::bgr888)
	{
		gl_format = GL_RGB;
		gl_type = GL_UNSIGNED_BYTE;
	}
	else if (src_format == CL_PixelFormat::abgr4444)
	{
		gl_format = GL_RGBA;
		gl_type = GL_UNSIGNED_SHORT_4_4_4_4_REV;
	}
	else if (src_format == CL_PixelFormat::bgr555)
	{
		gl_format = GL_RGB;
		gl_type = GL_UNSIGNED_SHORT_5_6_5_REV;
	}
	else
	{
		// Unsupported format by opengl
		gl_format = GL_RGBA;
		gl_type = GL_UNSIGNED_INT_8_8_8_8;
		do_pb_convert = true;
	}

	pixeldata.lock();

	if (do_pb_convert)
	{
		CL_PixelBuffer buffer(
			src_rect.get_width(),
			src_rect.get_height(),
			src_rect.get_width() * 4,
			CL_PixelFormat::rgba8888);

		buffer.lock();

		pixeldata.convert(
			buffer.get_data(),
			CL_PixelFormat::rgba8888,
			buffer.get_pitch(),
			CL_Rect(0, 0, src_rect.get_width(), src_rect.get_height()),
			src_rect);

		glTexSubImage2D(
			GL_TEXTURE_2D,
			0,
			pos.x,
			pos.y,
			src_rect.get_width(),
			src_rect.get_height(),
			gl_format,
			gl_type,
			buffer.get_data());

		buffer.unlock();
	}
	else
	{
		if (src_format.get_depth()*pixeldata.get_width()/8 != (int)pixeldata.get_pitch())
		{
			glPixelStorei(GL_UNPACK_ALIGNMENT, pixeldata.get_pitch());
		}
		glPixelStorei(GL_UNPACK_ROW_LENGTH, pixeldata.get_width());
		glPixelStorei(GL_UNPACK_SKIP_PIXELS, src_rect.left);
		glPixelStorei(GL_UNPACK_SKIP_ROWS, src_rect.top);

		glTexSubImage2D(
			GL_TEXTURE_2D,
			0,
			pos.x,
			pos.y,
			src_rect.get_width(),
			src_rect.get_height(),
			gl_format,
			gl_type,
			pixeldata.get_data());

		if (src_format.get_depth()*pixeldata.get_width()/8 != (int)pixeldata.get_pitch())
		{
			glPixelStorei(GL_UNPACK_ALIGNMENT, 0);
		}
		glPixelStorei(GL_UNPACK_ROW_LENGTH, -1);
		glPixelStorei(GL_UNPACK_SKIP_PIXELS, 0);
		glPixelStorei(GL_UNPACK_SKIP_ROWS, 0);
	}

	pixeldata.unlock();

	sync_pbuffer();
}

//: Draw surface on screen.
void CL_Surface_Target_OpenGL::draw(
	CL_Surface_Generic *attributes,
	int x, int y,
	CL_GraphicContext *gc)
{
	sync_texture();
	gc->impl->select_gc();

	// Calculate translation hotspot
	CL_Pointf translation = calc_hotspot(
		attributes->translation_origin,
		attributes->translation_hotspot,
		screen_size);

	init_attributes(attributes, screen_size);

	if (attributes->scale_x == 1.0f && attributes->scale_y == 1.0f)
	{
		float x_pos = x - translation.x;
		float y_pos = y - translation.y;
		state.rotation_hotspot.x += x_pos;
		state.rotation_hotspot.y += y_pos;

		CL_GLStateChecker::from_gc(gc)->add_quad(
			state,
			x_pos,
			y_pos,
			x_pos + static_cast<int>(screen_size.width),
			y_pos + static_cast<int>(screen_size.height),
			0,
			0,
			static_cast<int>(screen_size.width),
			static_cast<int>(screen_size.height));
	}
	else
	{
		float x_pos = x - translation.x * attributes->scale_x;
		float y_pos = y - translation.y * attributes->scale_y;
		state.rotation_hotspot.x = x_pos + state.rotation_hotspot.x * attributes->scale_x;
		state.rotation_hotspot.y = y_pos + state.rotation_hotspot.y * attributes->scale_y;

		CL_GLStateChecker::from_gc(gc)->add_quad(
			state,
			x_pos,
			y_pos,
			x_pos + screen_size.width * attributes->scale_x,
			y_pos + screen_size.height * attributes->scale_y,
			0,
			0,
			static_cast<int>(screen_size.width),
			static_cast<int>(screen_size.height));
	}
}

//: Draw surface on screen.
void CL_Surface_Target_OpenGL::draw(
	CL_Surface_Generic *attributes,
	const CL_Rect &dest,
	CL_GraphicContext *gc)
{
	sync_texture();
	gc->impl->select_gc();

	CL_Sizef size((float) dest.get_width(), (float) dest.get_height());
	
	// Calculate translation hotspot
	CL_Pointf translation = calc_hotspot(
		attributes->translation_origin,
		attributes->translation_hotspot,
		size);

//	CL_GLState state = init_attributes(attributes, size);
	init_attributes(attributes, size);

	if (attributes->scale_x == 1.0f && attributes->scale_y == 1.0f)
	{
		state.rotation_hotspot.x += dest.left-translation.x;
		state.rotation_hotspot.y += dest.top-translation.y;

		CL_GLStateChecker::from_gc(gc)->add_quad(
			state,
			dest.left-translation.x,
			dest.top-translation.y,
			dest.right-translation.x,
			dest.bottom-translation.y,
			0,
			0,
			dest.get_width(),
			dest.get_height());
	}
	else
	{
		translation.x *= attributes->scale_x;
		translation.y *= attributes->scale_y;

		state.rotation_hotspot.x = dest.left-translation.x + state.rotation_hotspot.x * attributes->scale_x;
		state.rotation_hotspot.y = dest.top-translation.y + state.rotation_hotspot.y * attributes->scale_y;

		CL_GLStateChecker::from_gc(gc)->add_quad(
			state,
			dest.left-translation.x,
			dest.top-translation.y,
			dest.left+(dest.right-dest.left)*attributes->scale_x-translation.x,
			dest.top+(dest.bottom-dest.top)*attributes->scale_y-translation.y,
			0,
			0,
			dest.get_width(),
			dest.get_height());
	}
}

//: Draw surface on screen.
void CL_Surface_Target_OpenGL::draw(
	CL_Surface_Generic *attributes,
	const CL_Rect &src,
	const CL_Rect &dest,
	CL_GraphicContext *gc)
{
	sync_texture();
	gc->impl->select_gc();

	CL_Sizef size((float)dest.get_width(), (float)dest.get_height());

	// Calculate translation hotspot
	CL_Pointf translation = calc_hotspot(
		attributes->translation_origin,
		attributes->translation_hotspot,
		size);

	init_attributes(attributes, size);

	if (attributes->scale_x == 1.0f && attributes->scale_y == 1.0f)
	{
		state.rotation_hotspot.x += dest.left-translation.x;
		state.rotation_hotspot.y += dest.top-translation.y;

		CL_GLStateChecker::from_gc(gc)->add_quad(
			state,
			dest.left-translation.x,
			dest.top-translation.y,
			dest.right-translation.x,
			dest.bottom-translation.y,
			src.left,
			src.top,
			src.right,
			src.bottom);
	}
	else
	{
		translation.x *= attributes->scale_x;
		translation.y *= attributes->scale_y;

		state.rotation_hotspot.x = dest.left-translation.x + state.rotation_hotspot.x * attributes->scale_x;
		state.rotation_hotspot.y = dest.top-translation.y + state.rotation_hotspot.y * attributes->scale_y;

		CL_GLStateChecker::from_gc(gc)->add_quad(
			state,
			dest.left-translation.x,
			dest.top-translation.y,
			dest.left+(dest.right-dest.left)*attributes->scale_x-translation.x,
			dest.top+(dest.bottom-dest.top)*attributes->scale_y-translation.y,
			src.left,
			src.top,
			src.right,
			src.bottom);
	}
}

void CL_Surface_Target_OpenGL::select_context(bool sync_buffer)
{
	if (sync_buffer)
	{
		sync_pbuffer();
		pbuffer_changed = true;
	}
	else
	{
#ifdef WIN32
		if (pbuffer_context == 0) return;
		CL_GLStateChecker::set_context(pbuffer_dc, pbuffer_context);

		// todo: check if wglQueryPbufferARB(evilness, WGL_PBUFFER_LOST_ARB, lazyhwcoders) is true.
		// Destroy and create a new pbuffer if this happened (weee!)
#else
		// todo: add linux pbuffer select context code here
#endif
	}
}

void CL_Surface_Target_OpenGL::sync_pbuffer()
{
	if (texture_changed && gc != 0)
	{
		select_context(false);

		// Copy texture to pbuffer:
		glEnable(GL_TEXTURE_2D);
		glBindTexture(GL_TEXTURE_2D, handle);
		glMatrixMode(GL_TEXTURE);
		glPushMatrix();
		GLdouble scale_matrix[16];
		memset(scale_matrix, 0, sizeof(GLdouble)*16);
		scale_matrix[0] = 1.0/texture_size.width;
		scale_matrix[5] = 1.0/texture_size.height;
		scale_matrix[10] = 1.0;
		scale_matrix[15] = 1.0;
		glLoadMatrixd(scale_matrix);
		glMatrixMode(GL_MODELVIEW);

		glDisable(GL_BLEND);
		glBegin(GL_QUADS);
		glColor3f(1.0f, 1.0f, 1.0f);
		glTexCoord2f(screen_size.width, 0.0f);
		glVertex2f(screen_size.width, 0.0f);
		glTexCoord2f(screen_size.width, screen_size.height);
		glVertex2f(screen_size.width, screen_size.height);
		glTexCoord2f(0.0f, screen_size.height);
		glVertex2f(0.0f, screen_size.height);
		glTexCoord2f(0.0f, 0.0f);
		glVertex2f(0.0f, 0.0f);
		glEnd();
		glEnable(GL_BLEND);

		glMatrixMode(GL_TEXTURE);
		glPopMatrix();
		glMatrixMode(GL_MODELVIEW);
		glDisable(GL_TEXTURE_2D);
	}
	texture_changed = false;
}

void CL_Surface_Target_OpenGL::sync_texture()
{
	if (pbuffer_changed && gc != 0)
	{
		select_context(false);

		// Copy pbuffer to texture:
		gc->flush();
		glEnable(GL_TEXTURE_2D);
		glBindTexture(GL_TEXTURE_2D, handle);
		glCopyTexSubImage2D(
			GL_TEXTURE_2D,
			0,
			0, 0,
			0, 0,
			int(screen_size.width), int(screen_size.height));
		glDisable(GL_TEXTURE_2D);
	}
	pbuffer_changed = false;
}

/////////////////////////////////////////////////////////////////////////////
// CL_Surface_Target_OpenGL implementation:

#ifdef WIN32
CL_Surface_Target_OpenGL::ptr_wglCreatePbufferARB CL_Surface_Target_OpenGL::wglCreatePbufferARB = 0;

CL_Surface_Target_OpenGL::ptr_wglGetPbufferDCARB CL_Surface_Target_OpenGL::wglGetPbufferDCARB = 0;

CL_Surface_Target_OpenGL::ptr_wglReleasePbufferDCARB CL_Surface_Target_OpenGL::wglReleasePbufferDCARB = 0;

CL_Surface_Target_OpenGL::ptr_wglDestroyPbufferARB CL_Surface_Target_OpenGL::wglDestroyPbufferARB = 0;

CL_Surface_Target_OpenGL::ptr_wglQueryPbufferARB CL_Surface_Target_OpenGL::wglQueryPbufferARB = 0;
#endif
