/*  $Id: menu_node_generic.cpp,v 1.3 2003/11/02 15:07:49 grumbel Exp $
**
**  ClanLib Game SDK
**  Copyright (C) 2003  The ClanLib Team
**  For a total list of contributers see the file CREDITS.
**
**  This library is free software; you can redistribute it and/or
**  modify it under the terms of the GNU Lesser General Public
**  License as published by the Free Software Foundation; either
**  version 2.1 of the License, or (at your option) any later version.
**
**  This library is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
**  Lesser General Public License for more details.
**
**  You should have received a copy of the GNU Lesser General Public
**  License along with this library; if not, write to the Free Software
**  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
**
*/

#include "precomp.h"
#include "menu_node_generic.h"
#include "menu_generic.h"
#include "API/GUI/menu.h"
#include "API/GUI/menu_node.h"
#include "API/signals.h"

/////////////////////////////////////////////////////////////////////////////
// Construction:

CL_MenuNode_Generic::CL_MenuNode_Generic(
	CL_MenuNode *self,
	CL_Menu *parent_menu)
:
	data(0),
	submenu(0),
	parent_menu(parent_menu),
	node(self)
{
	slots.connect( node->sig_mouse_enter(), this, &CL_MenuNode_Generic::on_mouse_enter);	
	slots.connect( node->sig_mouse_down(), this, &CL_MenuNode_Generic::on_mouse_down);
	slots.connect( node->sig_child_add(), this, &CL_MenuNode_Generic::on_child_add);
}

CL_MenuNode_Generic::~CL_MenuNode_Generic()
{
//	delete submenu;
}

/////////////////////////////////////////////////////////////////////////////
// Attributes:

bool CL_MenuNode_Generic::has_mouse_over()
{
	if( node->CL_Component::has_mouse_over() )
		return true;

	if( data )
		return data->has_mouse_over();
	
	return false;
}

bool CL_MenuNode_Generic::has_mouse_in_submenus()
{
	if( !submenu )
		return false;

	if( submenu->has_mouse_over() )
		return true;

	return submenu->has_mouse_in_submenus();
}

bool CL_MenuNode_Generic::has_open_submenu()
{
	if( submenu )
		return submenu->is_visible();

	return false;
}

/////////////////////////////////////////////////////////////////////////////
// Operations:

void CL_MenuNode_Generic::collapse()
{
	if( !submenu )
		return;

	submenu->collapse();
}

void CL_MenuNode_Generic::open_submenu()
{
	if( !submenu )
		return;

	submenu->open();
}

void CL_MenuNode_Generic::set_submenu(CL_Menu *menu)
{
	submenu = menu;
	submenu->set_parent_node(node);

//	parent_menu->get_root_menu()->reposition();
}

void CL_MenuNode_Generic::set_parent_menu(CL_Menu *menu)
{
	parent_menu = menu;

//	parent_menu->get_root_menu()->reposition();
}


/////////////////////////////////////////////////////////////////////////////
// Callbacks:

void CL_MenuNode_Generic::on_child_add(CL_Component *child)
{
	this->data = child;
	
	node->set_width(child->get_width());
	node->set_height(child->get_height());
	
	slots.connect( child->sig_mouse_enter(), this, &CL_MenuNode_Generic::on_mouse_enter );

	node->find_preferred_size();
}

void CL_MenuNode_Generic::on_mouse_down(const CL_InputEvent &e)
{
	if( !parent_menu->is_vertical() ) 
		parent_menu->set_click_to_open(false);

	if( data )
	{
		// if data is focusable, don't close menus when clicked.
		// This allows using dialogs in menu items.
		if( data->is_focusable() )
			return;
	}
	
	// click to open a submenu in a horizontal menu.
	if( submenu )
	{
		if( !submenu->is_open() )
		{
			open_submenu();
			node->get_parent_menu()->impl->close_lock = true; // todo: too hacky?
		}
		return;
	}
	
	// a menu item clicked, close menus.
	parent_menu->get_root_menu()->collapse_submenus();
}

void CL_MenuNode_Generic::on_mouse_enter()
{
	parent_menu->collapse_submenus();

	if( parent_menu->is_vertical() )
		open_submenu();
	else if( !parent_menu->is_click_to_open() ) // horizontal
		open_submenu();
}

